/*
** Module   : Broker.java
** Abstract : server side broker implementation.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 MAG 20140715 Implements remote launcher (broker).
** 002 GES 20150423 Fixed incorrect logging text.
*/

package com.goldencode.p2j.main;

import java.util.*;
import java.util.concurrent.*;
import java.util.logging.*;

import com.goldencode.p2j.net.Session;
import com.goldencode.p2j.util.*;

/**
 * Broker implementation. Holds brokers attributes read from directory as well as
 * parameters for remote broker instances linked to application server broker.  
 */
public class Broker
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(Broker.class.getName());

   /** List of accounts for this broker */
   private List<String> accounts;
   
   /** Remote broker instances registered for this broker */
   private Map<String, BrokerParameters> instances = new ConcurrentHashMap<>();

   /**
    * Constructor.
    * 
    * @param   accounts
    *          Accounts list or null if no accounts are defined for this broker.
    */
   public Broker(List<String> accounts)
   {
      this.accounts = accounts;
   }
   
   /**
    * Add an account for this broker.
    * 
    * @param   account
    *          P2J or OS account.
    */
   public void addAccount(String account)
   {
      accounts.add(account);
   }
   
   /**
    * Register a new remote broker instance.
    * 
    * @param   params
    *          Remote broker parameters.
    */
   public void addBroker(BrokerParameters params)
   {
      params.setUuid(UUID.randomUUID().toString());
      instances.put(params.getUuid(), params);
   }
   
   /**
    * Close brokers session on server shutdown.
    */
   public void shutdown()
   {
      synchronized (instances)
      {
         for (Map.Entry<String, BrokerParameters> entry : instances.entrySet())
         {     
            BrokerParameters params = entry.getValue();            
            if (params != null && params.getSession() != null)
            {
               params.getSession().terminate();
            }
         }
      }
   }

   /**
    * Remove a broker instance based on session.
    * 
    * @param   session
    *          Broker session.
    *          
    * @return  <code>true</code> on successful remove.
    */
   public boolean removeBroker(Session session)
   {
      synchronized (instances)
      {
         for (Map.Entry<String, BrokerParameters> entry : instances.entrySet())
         {     
            BrokerParameters params = entry.getValue();
            
            if (session != null && session.equals(params.getSession()))
            {
               params.close();               
               instances.remove(entry.getKey());
                            
               LOG.logp(Level.INFO,
                        "Broker.removeBroker()",
                        "",
                        LogHelper.generate("Broker user=%s uuid=%s removed.", 
                                           params.getUserId(), 
                                           params.getUuid()));               
               return true;
            }
         }
         
         return false;
      }
   }
   
   /**
    * Check if a specific account is define for this broker.
    * 
    * @param   account
    *          P2J or OS account.
    *          
    * @return  <code>true</code> if account is defined.
    */
   public boolean hasAccount(String account)
   {
      return accounts != null && accounts.contains(account);
   }
   
   /**
    * Check if broker has no accounts.
    * 
    * @return  <code>true</code> if no accounts are defined.
    */
   public boolean noAccounts()
   {
      return accounts == null || accounts.isEmpty();
   }

   /**
    * Get broker instances.
    * 
    * @return  A map containing the registered brokers.
    */
   public Map<String, BrokerParameters> getBrokers()
   {
      return instances;
   }

   /**
    * Find a broker based on UUID
    * 
    * @param   uuid
    *          Broker UUID.
    *          
    * @return  <code>true</code> when broker is found.
    */
   public boolean findBroker(String uuid)
   {
      return instances.containsKey(uuid);
   }
   
   /**
    * Get a broker instance based on UUID
    * 
    * @param   uuid
    *          Broker UUID.
    *          
    * @return  A broker instance.
    */
   public BrokerParameters getBroker(String uuid)
   {
      return instances.get(uuid);
   }
}
