/*
** Module   : WebClientBuilder.java
** Abstract : build and start a platform process for web chui client.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140206 First version, extracted from ClientBuilder.
** 002 MAG 20140711 Enable/Disable clipboard.
** 003 MAG 20140722 Implement remote launcher.
** 004 GES 20150312 Added support for the GUI web client.
*/

package com.goldencode.p2j.main;

import java.util.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.net.*;

// we explicitly import this class to avoid other references to the UI classes 
import com.goldencode.p2j.ui.client.OutputManager;

/**
 * This class is used to create and spawn a web client process using a {@link ProcessBuilder}
 * instance.  A special small tools are designed for Linux / Windows OS in order to spawn the 
 * process on the OS user account.  Both tools are written in C using specific OS API and are 
 * build as executable using ant native target.  The source code and the make file for this tools 
 * are available in the /src/native process directory.  The password is provided vis stdin 
 * redirection on both platforms.  The command line parameters for the spawned process are 
 * constructed from default parameters stored in directory and some specific parameters stored 
 * inside a {@link WebClientBuilderParameters} structure.
 */
public class WebClientBuilder
extends ClientBuilder
{   
   /** Server key store */
   private ServerKeyStore serverKeyStore;
   
   /**
    * Constructor.
    * 
    * @param   config
    *          The specific parameters for this client.
    */
   public WebClientBuilder(ClientBuilderParameters config)
   {
      super(config);

      BootstrapConfig serverConfig = SessionManager.get().config();
      
      // load server key store
      serverKeyStore = new ServerKeyStore(serverConfig);
   }
 
   /**
    * Get server key store
    * 
    * @return  The loaded server key store.
    */
   public ServerKeyStore getServerKeyStore()
   {
      return (serverKeyStore != null && serverKeyStore.isValid()) ? serverKeyStore : null;
   }

   /**
    * Add terminal options.
    */
   protected void addClientOptions()
   {
      WebClientBuilderParameters webConfig = (WebClientBuilderParameters) config;
      Map<String, String> options = webConfig.getClientOptions();
      
      boolean isGui = webConfig.isGui();

      // add to command
      if (isGui)
      {
         // GUI driver options
         command.add("client:driver:type=" + OutputManager.GUI_DRIVER_WEB);
      }
      else
      {
         // ChUI driver options
         command.add("client:driver:type=" + OutputManager.CHUI_DRIVER_WEB);
         command.add("client:chui:rows=" + options.get("rows"));
         command.add("client:chui:columns=" + options.get("columns"));
         command.add("client:chui:background=" + options.get("background"));
         command.add("client:chui:foreground=" + options.get("foreground"));
         command.add("client:chui:selection=" + options.get("selection"));
         command.add("client:chui:fontname=" + options.get("fontname"));
         command.add("client:chui:fontsize=" + options.get("fontsize"));
      }
      
      command.add("client:web:socketTimeout=" + options.get("webSocketTimeout"));
      command.add("client:web:watchdogTimeout=" + options.get("watchdogTimeout"));
      
      if (options.get("port") != null)
      {
         command.add("client:web:port=" + options.get("port"));
      }
      
      if (options.get("host") != null)
      {
         // do not send this parameter on remote spawn, let the remote web server select the IP
         if (!remote)
         {
            command.add("client:web:host=" + options.get("host"));            
         }
      }
      
      // add referrer address
      String referrer = webConfig.getReferrer();
      
      if (remote && referrer != null)
      {
         referrer = referrer.replace("localhost", BrokerCore.PARAM_HOST);
      }
      
      command.add("web:referrer:url=" + referrer);
      
      // clipboard enable/disable
      command.add("client:web:clipboard=" + options.get("clipboard"));
   }

   /**
    * Get the spawn arguments.  Web chui clients call the {@code spawn} tool using the
    * {@code spawn 0 <user> <workdir> <command> [args]} syntax.
    * <p>
    * As these arguments are already built into the {@link ClientBuilder#command} list, it just
    * returns the already build list with a "1" prefix. 
    * 
    * @return   See above.
    */
   @Override
   protected List<String> getSpawnArguments()
   {
      List<String> cmd = new ArrayList<>();
      
      // add mode
      cmd.add("1");
      
      // the entire command
      cmd.addAll(command);
      
      return cmd;
   }
}
