/*
** Module   : EmulatedTerminalState.java
** Abstract : core terminal simulator logic
**
** Copyright (c) 2011-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20150416 Created initial version by moving logic from AbstractSimulator/ChuiSimulator.
*/

package com.goldencode.p2j.ui.client.chui.driver;

import java.util.*;

import com.goldencode.p2j.ui.client.*;

/**
 * This class provides shared emulated terminal simulator logic. It manages the state for a
 * rectangular component that appears to look and act like a character terminal (TTY). However,
 * this class itself is not that component because it must be shared by different UI modalities.
 * For example, it will be used for both Swing and Web terminal emulators.
 * <p>
 * All output is stored in a 2-dimensional buffer of of cells, which store the character to be
 * drawn and the attributes/color of that character. 
 * <p>
 * The current cursor position is remembered and can be manipulated by the application. The
 * cursor visibility state can be externally controlled as well.
 * <p>
 * Input is read using key events and that is kept in a FIFO buffer as if they had been sent by
 * the remote terminal hardware.
 */
public abstract class EmulatedTerminalState
{
   /** Terminal configuration storage. */
   protected TerminalOptions options = null;
   
   /** Lock object for synchronizing configuration usage and changes. */
   protected Object cfgLock = new Object();
   
   /** Lock object for synchronizing selection processing. */
   protected Object selLock = new Object();
   
   /** Lock object for synchronizing data and cursor usage and changes. */
   protected Object dataLock = new Object();

   /** Flag that controls if the cursor is visible or not. */
   protected boolean showCursor = true;
   
   /** Screen contents. */
   protected Cell[][] data = null;
   
   /** Current X coordinate of the cursor. */
   protected int cursorX = 0;

   /** Current Y coordinate of the cursor. */
   protected int cursorY = 0;
   
   /**
    * Create a new instance.
    */
   public EmulatedTerminalState()
   {
   }
   
   /**
    * Copy the given cells into the 2-dimensional output buffer starting at
    * the current cursor position and writing to the "right", incrementing
    * the cursor X position for each cell. When the last column is reached,
    * the cursor X will wrap back to 0 and the cursor Y will increment to the
    * next row. When the last column of the last row is written, both the
    * cursor X and Y positions will wrap back to 0.
    * <p>
    * This write operation does not cause a repaint to occur, but the next
    * repaint that does occur will do so with the updated content.
    *
    * @param    output
    *           List of cells to write into the buffer.
    */
   public void append(Cell[] output)
   {
      int w = -1;
      int h = -1;
      
      synchronized (cfgLock)
      {
         w = options.columns;
         h = options.rows;
      }
      
      synchronized (dataLock)
      {
         for (Cell next : output)
         {
            // copy the data to the current cursor position and increment the
            // X cursor
            data[cursorY][cursorX++] = next;
            
            // wrap our cursor if needed
            if (cursorX == w)
            {
               cursorX = 0;
               cursorY++;
               
               if (cursorY == h)
               {
                  cursorY = 0;
               }
            }
         }
      }
   }
   
   /**
    * Completely replace the current 2-dimensional output buffer with the
    * content on the given <code>screen</code> parameter and then set the
    * the cursor position to the given values. Only non-null cells are copied.
    * Any cells that are null are ignored and left alone in the target buffer.
    * <p>
    * This write operation DOES cause a repaint to occur.
    *
    * @param    screen
    *           2-dimensional array of cells to write into the buffer. This
    *           must have the same dimensions as the output buffer and all
    *           rows must have column arrays of the same size.
    * @param    newX
    *           Cursor X position after the method call.
    * @param    newY
    *           Cursor Y position after the method call.
    */
   public void replace(Cell[][] screen, int newX, int newY)
   {
      int w = -1;
      int h = -1;
      
      synchronized (cfgLock)
      {
         w = options.columns;
         h = options.rows;
      }
      
      synchronized (dataLock)
      {
         // safety code
         if (screen.length == h && data.length == h)
         {
            for (int row = 0; row < h; row++)
            {
               // more safety code
               if (screen[row].length == w && data[row].length == w)
               {
                  for (int col = 0; col < w; col++)
                  {
                     // no need to clone each cell since the cell class' data
                     // is immutable
                     if (screen[row][col] != null)
                     {
                        data[row][col] = screen[row][col];
                     }
                  }
               }
            }
         }
         
         setCursorWorker(newX, newY, w, h);
      }
      
      triggerRepaint(true);
   }
   
   /**
    * Overwrite the entire 2-dimensional output buffer with the space character
    * to be rendered with the <code>Color.NORMAL</code> color. Sets the cursor
    * Y and X positions to 0.
    */
   public void clear()
   {
      int w = -1;
      int h = -1;
      
      synchronized (cfgLock)
      {
         w = options.columns;
         h = options.rows;
      }
      
      synchronized (dataLock)
      {
         for (int row = 0; row < h; row++)
         {
            for (int col = 0; col < w; col++)
            {
               data[row][col] = Cell.EMPTY_CELL;
            }
         }
         
         setCursorWorker(0, 0, w, h);
      }
   }
   
   /**
    * Position the cursor.
    *
    * @param    x
    *           New X cursor position.
    * @param    y
    *           New Y cursor position.
    *
    * @return  <code>true</code> if cursor positioning succeeded.
    */
   public boolean setCursor(int x, int y)
   {
      int w = -1;
      int h = -1;
      
      synchronized (cfgLock)
      {
         w = options.columns;
         h = options.rows;
      }
      
      return setCursorWorker(x, y, w, h);
   }
   
   /**
    * Set the cursor visibility.
    *
    * @param    on
    *           <code>true</code> to make the cursor visible (subsequen repaint
    *           operations will then cause the cursor to be drawn).
    *
    * @return   The previous cursor visiblity state.
    */
   public boolean setCursorStatus(boolean on)
   throws IllegalStateException
   {
      boolean old = false;
      
      synchronized (dataLock)
      {
         old = showCursor;
         showCursor = on;
      }
      
      return old;
   }
   
   /**
    * Get the cursor's current X coordinate.
    *
    * @return   The current X coordinate.
    */
   public int getCursorX()
   {
      int x = -1;
      
      synchronized (dataLock)
      {
         x = cursorX;
      }
      
      return x;
   }
   
   /**
    * Get the cursor's current Y coordinate.
    *
    * @return   The current Y coordinate.
    */
   public int getCursorY()
   {
      int y = -1;
      
      synchronized (dataLock)
      {
         y = cursorY;
      }
      
      return y;
   }
   
   /**                                         
    * Gets the number of rows on the screen.
    *
    * @return   The number of rows (lines of characters).
    */
   public int getRows()
   {
      int r = -1;
      
      synchronized (cfgLock)
      {
         r = options.rows;
      }
      
      return r;
   }

   /**
    * Gets the number of columns on the screen.
    *
    * @return   The number of columns (in characters).
    */
   public int getColumns()
   {
      int c = -1;
      
      synchronized (cfgLock)
      {
         c = options.columns;
      }
      
      return c;
   }

   /**
    * Initialize the instance using the given configuration. This will obtain
    * and initialize its fonts (and the color mapper). It will use the font
    * metrics to calculate the size of the component (based on the number of
    * rows and columns).  If this is the first initialization, the 2-dimensional
    * output buffer will be created and cleared.
    *
    * @param    terminalOptions
    *           The configuration for the terminal simulator.
    */
   public void configure(TerminalOptions terminalOptions)                    
   {
      int w = -1;
      int h = -1;
      
      synchronized (cfgLock)
      {
         this.options = (terminalOptions == null) ? new TerminalOptions() : terminalOptions;
         
         w = options.columns;
         h = options.rows;
      }
      
      synchronized (dataLock)
      {
         if (data == null)
         {
            data = new Cell[h][w];
            clear();
         }
      }
   }
   
   /**
    * Cause a partial or full repaint to occur. If <code>full</code> is
    * <code>true</code> or if there is no current selection operation occurring,
    * a repaint of the entire component ("screen") will occur.  Otherwise,
    * a repaint of only the selection area will occur.
    *
    * @param    full
    *           Flag to force a the entire screen to repaint.
    */
   public abstract void triggerRepaint(boolean full);
   
   /**
    * Get screen memory array.
    * 
    * @return  The screen memory array.
    */
   public Cell[][] getScreenData()
   {
      return data;
   }

   /**
    * Get terminal options used by client.
    * 
    * @return  Terminal options.
    */
   public TerminalOptions getTerminalOptions()
   {
      return this.options;
   }
   
   /**
    * Position the cursor.
    *
    * @param    x
    *           New X cursor position.
    * @param    y
    *           New Y cursor position.
    * @param    w
    *           Screen width.
    * @param    h
    *           Screen height.
    *
    * @return  <code>true</code> if cursor positioning succeeded.
    */
   private boolean setCursorWorker(int x, int y, int w, int h)
   {
      if (x < 0 || x >= w || y < 0 || y >= h)
         return false;
      
      synchronized (dataLock)
      {
         cursorX = x;
         cursorY = y;
      }
      
      return true;
   }
}
