/*
** Module   : BatchPrimitives.java
** Abstract : Output primitives for batch mode driver.
**
** Copyright (c) 2013-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20131104 Created initial version.
** 002 EVL 20140225 The DISPLAY statement requires special processing for screen update in Linux
**                  redirected case or Winodws background mode.  We have to display data on the
**                  current STDOUT.
** 003 GES 20150506 Modifications to match new common driver parent hierarchy.
*/

package com.goldencode.p2j.ui.client.chui.driver.batch;

import java.util.List;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.chui.driver.console.*;
import com.goldencode.p2j.util.*;
import com.goldencode.util.*;

/**
 * Implementation of the primitives for batch mode screen driver.
 * <p>
 * This implementation uses console-based null driver for terminal access.  This means all i/o
 * requests will produce the error messages and immediate stop for client application.  There are
 * two modes of functionality.  If batch mode is active the primitives generate error otherwise
 * the control is passing to the underlying console primitive handlers. 
 */
public class BatchPrimitives
extends ConsolePrimitives
{
   /** Flag indicating whether we need to do special processing or just ignore calls. */
   private boolean isActive = false;

   /**
    * Constructor.
    */
   public BatchPrimitives()
   {
   }

   /**
    * Activates special processing for batch mode.  This means all i/o terminal requests produce
    * error writing and immediate application stop.
    */
   public void activate()
   {
      isActive = true;
   }
   
   /**
    * Deactivates special processing for batch mode.  This means all i/o terminal requests will
    * be ignored and application will contimue.
    */
   public void deActivate()
   {
      isActive = false;
   }
   
   /**
    * Check if the special internal processing for batch mode is active.
    *
    * @return  <code>true</code> if batch mode is active, <code>false</code> otherwise.
    */
   public boolean isActive()
   {
      return isActive; 
   }
   
   /**
    * Position the virtual cursor during the drawing.
    *
    * @param   x
    *          New cursor X.
    * @param   y
    *          New cursor Y.
    *
    * @return  <code>true</code> if cursor positioning succeeded.
    */
   @Override
   public boolean cursorAt(int x, int y)
   {
      // At this ime we just do nothing here.
      return true;
   }

   /**
    * Add a string to the virtual terminal at the current cursor position.
    *
    * @param   str
    *          The text to output.
    * @param   color
    *          The Color instance used to draw characters. Mapping into native
    *          colors is performed internally.
    */
   @Override
   public void append(String str, Color color)
   {
      if (isActive)
      {
         // This is required to suppress any further error messages for i/o
         isActive = false;
         
         // Display error message and leave the application.
         BatchHelper.outputOperation();
      }
   }

   /**
    * Gets the number of rows on the screen.  This is artificial value for batch mode.  Windows
    * HEIGHT attribute returns the value that looks like desktop height in chracter units, Linux
    * version returns unknown value(?).  However we need the reasonable value to corectly
    * allocate internal character cell bitmaps. 
    *
    * @return  The number of rows (lines of characters) - 24 for batch mode.
    */
   @Override
   public int screenHeight()
   {
      // Note this is artificial value.
      return 24;
   }

   /**
    * Gets the number of columns on the screen.  This is artificial value for batch mode.  Windows
    * WIDTH attribute returns the value that looks like desktop height in chracter units, Linux
    * version returns unknown value(?).  However we need the reasonable value to corectly
    * allocate internal character cell bitmaps.
    *
    * @return  The number of columns (in characters).
    */
   @Override
   public int screenWidth()
   {
      // TODO: Find out what is the real value in batch mode here.
      return 80;
   }

   /**
    * Synchronize the state of the physical terminal with the state of
    * the virtual terminal.
    */
   @Override
   public void sync()
   {
      if (isActive)
      {
         // This is required to suppress any further error messages for i/o
         isActive = false;
         
         // Display error message and leave the application.
         BatchHelper.outputOperation();
      }
   }

   /**
    * Perform a bulk output operation for the whole screen at once with
    * output masking when desired. Set the cursor afterwards.
    * <p>
    * The screen container is safe for use with characters that are UNICODE
    * encoded which are also outside of the 7-bit ASCII range. Such
    * characters are passed through to the native layer directly and are
    * not converted to a natively encoded form.  This could be provided
    * but since one of the primary uses of this method is for drawing
    * characters (e.g. <code>ACS_HLINE</code>) which are encoded using the
    * full 32-bits, any attempt to convert these into a single byte native
    * encoding would lose data.
    * <p>
    * Because of the above limitation, DO NOT USE THIS FOR OUTPUT OF USER
    * DATA THAT COULD CONTAIN DATA OUTSIDE OF THE 7-bit ASCII RANGE.
    *
    * @param   screen
    *          Container for whole screen content and cursor position.
    */
   @Override
   public void updateScreen(ScreenData screen)
   {
      if (isActive)
      {
         // Special processing for redirected STDOUT and Windows background mode
         if (FileChecker.isConsoleRedirected() ||
             (PlatformHelper.isUnderWindowsFamily() && OutputManager.isBatchInBackground()))
         {
            Cell[][] data = screen.asArray();
            for (int i = 0; i < data.length; i++)
            {
               // Handle line by line
               StringBuffer sb = new StringBuffer();
               for (int j = 0; j < data[i].length; j++)
               {
                  if (data[i][j] != null)
                  {
                     char cellChar = data[i][j].character();
                     // Skip frame border line
                     if (cellChar == '\u250C' || cellChar == '\u2514')
                     {
                        break;
                     }

                     if (cellChar == '\u2500') // Substitute horizontal line
                     {
                        sb.append("-");
                     }
                     else if (cellChar != '\u2502') // Skip vertical frame border lines
                     {
                        character unicodeChar = new character(cellChar);
                        sb.append(unicodeChar.toStringMessage());
                     }
                  }
                  else
                  {
                     sb.append(" "); // Put space instead null cell
                  }
               }
               // Remove any unnecessary spase chars
               String lineOut = sb.toString().trim();
               // Output to STDOUT only if we have something to output
               if (lineOut.length() > 0)
               {
                  ErrorManager.displayError(lineOut);
               }
            }
         }
         else
         {
            // This is required to suppress any further error messages for i/o
            isActive = false;
            
            // Display error message and leave the application.
            BatchHelper.outputOperation();
         }
      }
   }
}
