/*
** Module   : SettingsPanel.java
** Abstract : terminal options editor panel that can be hosted in a dialog
**
** Copyright (c) 2011-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description--------------------------------
** 001 GES 20110921 Created initial version.
** 002 GES 20150420 Color data inputs are processed to/from ColorRgb now.
*/

package com.goldencode.p2j.ui.client.chui.driver.swing;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.*;

import com.goldencode.p2j.ui.client.driver.swing.*;
import com.goldencode.p2j.ui.client.chui.driver.*;

/**
 * Provides a terminal options editor panel provided as a modal dialog.
 */
public class SettingsPanel
extends JPanel
{
   /** Dialog title for the this panel. */
   public static final String SETTINGS_TITLE = "Edit Settings";
   
   /** Label text for the rows field. */
   public static final String ROWS_TXT = "Rows";
   
   /** Label text for the columns field. */
   public static final String COLS_TXT = "Columns";
   
   /** Label text for the font name field. */
   public static final String FN_TXT = "Font Name";
   
   /** Label text for the font size field. */
   public static final String FS_TXT = "Font Size";
   
   /** Label text for the background color button. */ 
   public static final String BGC_TXT = "Background Color";
   
   /** Label text for the foreground color button. */ 
   public static final String FGC_TXT = "Foreground Color";
   
   /** Label text for the selection color button. */ 
   public static final String SELC_TXT = "Selection Color";
   
   /** Configuration being edited. */
   private TerminalOptions options = null;
   
   /** Rows field. */
   private JFormattedTextField rows = null;
   
   /** Columns field. */
   private JFormattedTextField columns = null;
   
   /** Font name field. */
   private JSpinner fontname = null;
   
   /** Font size field. */
   private JFormattedTextField fontsize = null;
   
   /** Background color button. */
   private JButton background = null;
   
   /** Foreground color button. */
   private JButton foreground = null;
   
   /** Selection color button. */
   private JButton selection = null;
   
   /**
    * Create a new instance to edit the given terminal options.
    *
    * @param    options
    *           The configuration to edit.
    */
   private SettingsPanel(TerminalOptions options)
   {
      super(new SpringLayout(), true);
      this.options = options;
      
      JComponent[][] cells  = new JComponent[7][2];
      
      Color bg  = SwingHelper.fromColorRgb(options.background);
      Color fg  = SwingHelper.fromColorRgb(options.foreground);
      Color sel = SwingHelper.fromColorRgb(options.selection);
      
      rows       = addIntegerEditor(ROWS_TXT, options.rows, cells, 0);
      columns    = addIntegerEditor(COLS_TXT, options.columns, cells, 1);
      fontname   = addFontSpinner(FN_TXT, options.fontname, cells, 2);
      fontsize   = addIntegerEditor(FS_TXT, options.fontsize, cells, 3);
      background = addColorButton(BGC_TXT, bg, cells, 4);
      foreground = addColorButton(FGC_TXT, fg, cells, 5);
      selection  = addColorButton(SELC_TXT, sel, cells, 6);

      SwingHelper.setupGridSprings(this, cells, 5);
   }
   
   /**
    * Create a new instance of the settings panel class, display it as a
    * modal dialog box allowing edits and return the resulting edited options.
    *
    * @param    frame
    *           The parent frame.
    * @param    options
    *           The configuration to edit.
    *
    * @return   A new options instance with the modified values if the editing
    *           was accepted or the original unmodified options instance if
    *           the editing dialog was cancelled by the user.
    */
   public static TerminalOptions edit(Frame frame, TerminalOptions options)
   {
      SettingsPanel panel = new SettingsPanel(options);
      
      int rc = JOptionPane.showOptionDialog(frame,
                                            panel,
                                            SETTINGS_TITLE,
                                            JOptionPane.OK_CANCEL_OPTION,
                                            JOptionPane.PLAIN_MESSAGE,
                                            null,
                                            null,
                                            null);
      if (rc == 0)
      {
         return panel.getEdits();
      }
      
      // return the original instance on failure
      return options;
   }
   
   /**
    * Obtain any changes made in the editor. This is safe to be unsynchronized
    * as long as it is private and it is only called when the user is done with
    * any editing!
    *
    * @return   A new terminal options instance with its values set from the
    *           original instance (by default) or with edited values on a
    *           member-by-member basis.
    */
   private TerminalOptions getEdits()
   {
      TerminalOptions edits = new TerminalOptions();
      
      Integer rval = (Integer) rows.getValue();
      edits.rows = (rval == null) ? options.rows : rval;
      
      Integer cval = (Integer) columns.getValue();
      edits.columns = (cval == null) ? options.columns : cval;
      
      String fname = (String) fontname.getValue();
      edits.fontname = (fname == null) ? options.fontname : fname;
      
      Integer fsz = (Integer) fontsize.getValue();
      edits.fontsize = (fsz == null) ? options.fontsize : fsz;
      
      // the color that is displayed as the background of the button is the
      // associated value
      edits.background = SwingHelper.toColorRgb((Color) background.getBackground());
      edits.foreground = SwingHelper.toColorRgb((Color) foreground.getBackground());
      edits.selection  = SwingHelper.toColorRgb((Color) selection.getBackground());
      
      return edits;
   }
   
   /**
    * Add a JLabel and JFormattedTextField control for integer editing, into
    * the specified row of a 2 column layout of controls.
    *
    * @param    descr
    *           The text of the label which will be in the first column.
    * @param    val
    *           The default value for the field.
    * @param    cells
    *           The two-dimensional array of components to edit.
    * @param    row
    *           The row in which to insert the controls.
    *
    * @return   The text field control that was added.
    */
   private JFormattedTextField addIntegerEditor(String         descr,
                                                Integer        val,
                                                JComponent[][] cells,
                                                int            row)
   {
      JLabel label = new JLabel(descr, JLabel.RIGHT);
      add(label);
         
      JFormattedTextField text = new JFormattedTextField();
      text.setValue(val);
      text.setColumns(10);
      text.setHorizontalAlignment(JTextField.RIGHT);
      label.setLabelFor(text);
      add(text);
      
      cells[row][0] = label;
      cells[row][1] = text;
      
      return text;
   }
   
   /**
    * Add a JLabel and JButton control for color selection, into the specified
    * row of a 2 column layout of controls. The background color of the button
    * will be set to the given color so that the button natually shows the
    * currently selected color. The button's action listener will be set to
    * a new instance which will invoke a <code>JColorChooser</code> dialog.
    *
    * @param    descr
    *           The text of the label which will be in the first column.
    * @param    val
    *           The background color for the button.
    * @param    cells
    *           The two-dimensional array of components to edit.
    * @param    row
    *           The row in which to insert the controls.
    *
    * @return   The button control that was added.
    */
   private JButton addColorButton(final String   descr,
                                  final Color    val,
                                  JComponent[][] cells,
                                  int            row)
   {
      JLabel label = new JLabel(descr, JLabel.RIGHT);
      add(label);
         
      final JButton clr = new JButton();
      clr.setBackground(val);
      label.setLabelFor(clr);
      add(clr);
      
      clr.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent evt) 
         {
            Color newColor = JColorChooser.showDialog(SettingsPanel.this,
                                                      "Choose " + descr,
                                                      clr.getBackground());
            if (newColor != null)
               clr.setBackground(newColor);
         }
      });
      
      cells[row][0] = label;
      cells[row][1] = clr;
      
      return clr;
   }
   
   /**
    * Add a JLabel and JSpinner control for the fonts list, into the specified
    * row of a 2 column layout of controls. The size of the spinner will be as
    * wide as necessary to accomodate the largest font name.
    *
    * @param    descr
    *           The text of the label which will be in the first column.
    * @param    val
    *           The default value of the spinner.
    * @param    cells
    *           The two-dimensional array of components to edit.
    * @param    row
    *           The row in which to insert the controls.
    *
    * @return   The spinner control that was added.
    */
   private JSpinner addFontSpinner(String         descr,
                                   String         val,
                                   JComponent[][] cells,
                                   int            row)
   {
      JLabel label = new JLabel(descr, JLabel.RIGHT);
      add(label);
         
      String[]         flist   = SwingHelper.getMonoFontList(this);
      SpinnerListModel fonts   = new SpinnerListModel(flist);
      JSpinner         spinner = new JSpinner(fonts);
      spinner.setValue(val);
      
      // calculate size for the spinner based on the largest string
      FontMetrics fm  = spinner.getFontMetrics(spinner.getFont());
      int         wid = fm.stringWidth(largestFontName());
      Dimension   dim = new Dimension(wid, spinner.getHeight());
      spinner.setMinimumSize(dim);
      spinner.setMaximumSize(dim);
      spinner.setPreferredSize(dim);
      
      label.setLabelFor(spinner);
      add(spinner);
      
      cells[row][0] = label;
      cells[row][1] = spinner;
      
      return spinner;
   }
   
   /**
    * Helper to calculate which of the font names is the largest.
    *
    * @return   The largest font name or <code>null</code> if the font list
    *           is empty.
    */
   private String largestFontName()
   {
      String[] list = SwingHelper.getMonoFontList(this);
      String   big  = null;
      
      for (String font : list)
      {
         if (big == null || big.length() < font.length())
         {
            big = font;
         }
      }
      
      return big;
   }
}
