/*
** Module   : WebClientProtocol.java
** Abstract : provides the core WebSocket functionality for web clients
**
** Copyright (c) 2013-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 GES 20150317 Moved code from the ChuiWebSimulator and enhanced it.
*/

package com.goldencode.p2j.ui.client.driver.web;

import java.io.*;
import java.nio.*;
import java.util.*;
import java.util.concurrent.*;

import org.eclipse.jetty.websocket.api.*;
import org.eclipse.jetty.websocket.api.annotations.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;

/**
 * Implements the WebSockets protocol to drive communications with a JavaScript web client.
 * <p>
 * IMPORTANT: Do not use logging anywhere inside this class. When an authentication plugin
 * is running during session establishment with the P2J server using logging no messages are 
 * written into the log and the application is blocked.
 * <p>   
 * The payloads are in 2 formats binary and JSON text.
 * <p>
 * Input messages:
 * 
 * Keyboard input.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x01  message id
 * 1       uu    keystroke int value
 * 2       ll    keystroke = uu << 8 | ll 
 * <p>
 * Keyboard pair input.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x02  message id
 * 1       uu    keyCode int value
 * 2       ll    keyCode = uu << 8 | ll 
 * 3       uu    charCode int value
 * 4       ll    charCode = uu << 8 | ll 
 * <p>
 * Paste.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x03  message id
 * 1       uu    charCode int value
 * 2       ll    charCode = uu << 8 | ll 
 * ...
 * n-1     uu    charCode int value
 * n       ll    charCode = uu << 8 | ll 
 * <p>
 * Output messages:
 * 
 * Clear screen.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x80  message id
 * <p>
 * Draw text on screen.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x81  message id
 * 1..n    data  data packets
 * 
 * structure of data packets:

 * offset  value  content
 * ------------------------------------
 * 0       size   packet size high byte
 * 1       size   packet size low byte
 * 2       width  rectangle with
 * 3       height rectangle height
 * 4       row    row number
 * 5       col    column number
 * 6       attr   text attributes
 * 7       color  FG/BG color pairs
 * 8..n    text   text characters, each character on 2 bytes (UTF-16)
 * <p>
 * Set cursor position.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x82  message id
 * 1       row   row
 * 2       col   column
 * <p>
 * Set cursor status.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x83  message id
 * 1       stat  status 1 = visible 
 *                      0 = hidden 
 * <p>
 * Play a beep message.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x84  message id
 * <p>
 * Quit message.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x85  message id
 * <p>
 * Palette message is in JSON format as follow.
 * 
 *    {"c":0, "p":[color1, color2, ... color3]}
 * 
 * Example:
 *    {"c":0, "p":["#000000", "#FF0000", "#00FF00", "#0000FF", "#FFFFFF"]}
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x86  message id
 * 1       mode  mode 1 = VT100 
 *                    0 = P2J 
 * <p>
 * Switch mode.
 * 
 */
@SuppressWarnings("serial")
@WebSocket(maxIdleTime = 900000, maxTextMessageSize = 4096, maxBinaryMessageSize = 32768)
public class WebClientProtocol
implements WebClientMessageTypes
{
   /** Web socket timeout. */
   private final long timeout;
   
   /** Watchdog thread time out value. */
   private final long wdtimeout;
   
   /** Keyboard input queue. */
   private Deque<Integer> keyBuffer = new ConcurrentLinkedDeque<>();
   
   /** Watchdog thread instance. */
   private WatchdogTimer watchdog;
   
   /** Push messages worker thread. */
   private PushMessagesWorker pushWorker;
   
   /** Number of simultaneous connections. */
   private volatile int connections = 0;
   
   /** Keyboard input is being processed in virtual VT100 terminal mode. */
   private boolean isVT100 = false;
   
   /** Callbacks for delegated processing. */
   private ClientProtocolHooks callbacks = null;
         
   /** Instance lock for synchronization of operations. */
   protected final Object lock;

   /**
    * Create a instance.
    *
    * @param    lock
    *           Object to use for locking. By passing this in, the caller can share the
    *           locking instance with this code.  Use <code>null</code> if an internal
    *           lock object should be created.
    * @param    callbacks
    *           An instance with hooks to call at specific events.
    * @param    timeout
    *           Web socket timeout.
    * @param    wdtimeout
    *           Watchdog thread timeout.
    */
   public WebClientProtocol(Object              lock,
                            ClientProtocolHooks callbacks,
                            long                timeout,
                            long                wdtimeout)
   {
      this.lock      = (lock != null) ? lock : new Object();
      this.callbacks = callbacks;
      this.timeout   = timeout;
      this.wdtimeout = wdtimeout;
      
      startWatchdog(wdtimeout);
   }
   
   /**
    * This method is called when the remote peer open a web socket connection.
    *
    * @param   session
    *          An active link of communications with a Remote WebSocket Endpoint.
    *
    */
   @OnWebSocketConnect
   public void onConnect(Session session)
   {
      synchronized (lock)
      {
         if (++connections == 1)
         {
            // incoming connection means the watchdog is no longer needed
            stopWatchdog();
            
            if (timeout > 0)
            {
               session.setIdleTimeout(timeout);
            }
            
            startPushWorker(session);
            
            if (callbacks != null)
            {
               // synchronize the remove client with the current or initial state
               callbacks.initRemoteClient();
            }
         }
         else
         {
            // multiple connections are not allowed 
            session.close();
         }
      }
   }

   /**
    * This method is called when a text message is received.
    *
    * @param   message
    *          The text message.
    */
   @OnWebSocketMessage
   public void onMessage(String message) 
   {
      // not used
   }

   /**
    * This method is called when a binary message is received.
    *
    * @param   message
    *          Message content as an byte array.
    * @param   offset
    *          Message offset..
    * @param   length
    *          Message size.
    */
   @OnWebSocketMessage
   public void onMessage(byte[] message, int offset, int length)
   {
      if (length == 3 && message[offset] == WebClientProtocol.MSG_KEY)
      {
         int keystroke = (message[offset + 1] << 8) & 0xFF00;
         keystroke |= (message[offset + 2] & 0x00FF);
         
         synchronized (lock)
         {
            keyBuffer.offerLast(keystroke);
            lock.notifyAll();
         }
      }
      else if (length == 5 && message[offset] == WebClientProtocol.MSG_KEY_VT100)
      {
         int keyCode = (message[offset + 1] << 8) & 0xFF00;
         keyCode |= (message[offset + 2] & 0x00FF);
         
         int charCode = (message[offset + 3] << 8) & 0xFF00;
         charCode |= (message[offset + 4] & 0x00FF);
         
         callbacks.injectVT100Key(keyCode, (char) charCode);
      }
      else if (length > 1 && message[offset] == WebClientProtocol.MSG_PASTE)
      {
         for (int i = offset + 1;i < length;i += 2)
         {
            int charCode = (message[i] << 8) & 0xFF00;
            charCode |= (message[i + 1] & 0x00FF);
            
            if (isVT100)
            {
               callbacks.injectVT100Key(charCode, (char) charCode);
            }
            else
            {
               synchronized (lock)
               {
                  keyBuffer.offerLast(charCode);
                  lock.notifyAll();
               }
            }
         }
      }
   }

   /**
    * This method is called when the connection is closed.
    *
    * @param   statusCode
    *          The status integer code.
    * @param   reason
    *          A string representing the reason.
    */
   @OnWebSocketClose
   public void onClose(int statusCode, String reason)
   {
      synchronized (lock)
      {
         if (--connections == 0)
         {
            stopPushWorker();
            startWatchdog(wdtimeout);            
         }
      }
   }
   
   /**
    * This method is called on communication errors.
    *
    * @param   session
    *          An active link of communications with a Remote WebSocket Endpoint.
    * @param   error
    *          A <code>Throwable</code> instance representing the error condition.
    */
   @OnWebSocketError
   public void onError(Session session, Throwable error)
   {
      synchronized (lock)
      {
         session.close();      
      }
   }

   /**
    * Send a text message.
    * 
    * @param    message
    *           The fully encoded (JSON) text message.
    */   
   public void sendTextMessage(String message) 
   {
      synchronized (lock)
      {
         if (pushWorker != null)
         {
            pushWorker.pushMessage(message);
         }
      }
   }

   /**
    * Send a binary message of arbitrary size.
    * 
    * @param    message
    *           The fully encoded binary message contents.
    */   
   public void sendBinaryMessage(byte[] message) 
   {
      synchronized (lock)
      {
         if (pushWorker != null)
         {
            pushWorker.pushMessage(ByteBuffer.wrap(message));
         }
      }
   }

   /**
    * Convenience method to send a single-byte binary message.  The core functionality is
    * implemented in {@link #sendBinaryMessage(byte[])}.
    * 
    * @param    message
    *           The byte which comprises the entire binary message.
    */   
   public void sendBinaryMessage(byte message) 
   {
      sendBinaryMessage(new byte[] { message });
   }

   /**
    * Convenience method to send a two-byte binary message.  The core functionality is
    * implemented in {@link #sendBinaryMessage(byte[])}.
    * 
    * @param    first
    *           The first byte of the binary message.
    * @param    second
    *           The second byte of the binary message.
    */   
   public void sendBinaryMessage(byte first, byte second) 
   {
      sendBinaryMessage(new byte[] { first, second });
   }

   /**
    * Convenience method to send a three-byte binary message.  The core functionality is
    * implemented in {@link #sendBinaryMessage(byte[])}.
    * 
    * @param    first
    *           The first byte of the binary message.
    * @param    second
    *           The second byte of the binary message.
    * @param    third
    *           The third byte of the binary message.
    */   
   public void sendBinaryMessage(byte first, byte second, byte third) 
   {
      sendBinaryMessage(new byte[] { first, second, third });
   }

   /**
    * Read the next key from the FIFO key input buffer.  
    *
    * @return   The next key in the buffer or -1 if buffer is empty.
    */
   public int readKey()
   {
      synchronized (lock)
      {
         int result = -1;
         
         while (keyBuffer.isEmpty())
         {
            try
            {
               lock.wait();
            }
            catch (InterruptedException e)
            {
               break;
            }
         }

         Integer key = keyBuffer.pollFirst();
         
         if (key != null)
         {
            result = key.intValue();
         }

         return result;
      }
   }
   
   /**
    * The beep() should be implemented by sending a command on the remote end point.
    * The command is part of the communication protocol.
    *
    */
   public void beep()
   {
      sendBinaryMessage(MSG_BEEP);         
   }
   
   /**
    * Remote application is about to end. Notify the JS client about that.
    * The client should be redirected to referrer address.
    */
   public void quit()
   {
      sendBinaryMessage(MSG_QUIT);         
   }

   /**
    * Switch between P2J client and interactive client.
    * 
    * @param    isVT100
    *           If <code>true</code>, process keys in virtual VT100 terminal mode otherwise
    *           process for internal client screen.
    */
   public void swapKeyboardMode(boolean isVT100)
   {
      synchronized (lock)
      {
         this.isVT100 = isVT100;
      }
      sendSwitchMode();
   }   

   /**
    * Switch mode message.
    */
   public void sendSwitchMode()
   {
      sendBinaryMessage(MSG_VT100, (byte)(this.isVT100 ? 1 : 0));               
   }
   
   /**
    * Start watchdog timer.
    * <p>
    * This must only be called by a caller that owns the instance lock.
    * 
    * @param    wdtimeout
    *           Watchdog thread timeout value.
    */
   private void startWatchdog(long wdtimeout)
   {
      if (watchdog == null)
      {
         watchdog = new WatchdogTimer(wdtimeout);
      }
   }

   /**
    * Stop watchdog timer.
    * <p>
    * This must only be called by a caller that owns the instance lock.
    */
   private void stopWatchdog()
   {
      if (watchdog != null)
      {
         watchdog.kill();
         watchdog = null;
      }
   }

   /**
    * Start pushWorker thread.
    * <p>
    * This must only be called by a caller that owns the instance lock.
    * 
    * @param   session
    *          An active link of communications with a Remote WebSocket end point.
    */
   private void startPushWorker(Session session)
   {
      if (pushWorker == null)
      {
         pushWorker = new PushMessagesWorker(session);
      }
   }

   /**
    * Stop pushWorker thread.
    * <p>
    * This must only be called by a caller that owns the instance lock.
    */
   private void stopPushWorker()
   {
      if (pushWorker != null)
      {
         pushWorker.stopPushing();
         pushWorker = null;
      }
   }
}