/*
** Module   : GuiFontResolver.java
** Abstract : GUI widget font resolver
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20141022 Created initial version.
** 002 CA  20141220 When working with fonts, these need to be scaled to match their legacy width
**                  and height.
** 003 CA  20150317 If a font is not specified, resolve it from the default-font or 
**                  default-fixed-font, depending if the widget requires a fixed font or not. 
** 004 CA  20150406 If undefined font, get it from the frame (if exists).
** 005 CA  20150507 No need to re-scale the font, it gets scaled when it gets created - so all 
**                  fonts are assumed to match the legacy metrics.
** 006 GES 20150425 Eliminated AWT dependencies.
*/

package com.goldencode.p2j.ui.client.gui;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;

/**
 * Resolve widget font. 
 */
public class GuiFontResolver
{
   /** <code>true</code> when there is a newly added font. */
   public boolean fontChanged;
   
   /** Widget font cache. */
   private Map<Integer, FontCache> fontCache = new HashMap<>();

   /** Window instance. */
   private Window<?> window;
      
   /** Screen driver. */
   private GuiDriver gd;
   
   /** Widget config. */
   private BaseConfig config;

   /**
    * Flag identifying that the referenced font must default to the default fixed font, if is 
    * not specified.
    */
   private boolean fixedFont;

   /**
    * Constructor.
    * <p>
    * If the referenced {@link BaseConfig#font} is undefined, it defaults to the 
    * {@link FontManager#DEFAULT_FONT}.
    * 
    * @param    window
    *           Window instance.
    * @param    gd
    *           Screen driver.
    * @param    config
    *           Widget configuration.
    */
   public GuiFontResolver(Window<?>  window, GuiDriver gd, BaseConfig config)
   {
      this(window, gd, config, false);
   }

   /**
    * Constructor.
    * 
    * @param    window
    *           Window instance.
    * @param    gd
    *           Screen driver.
    * @param    config
    *           Widget configuration.
    * @param    fixedFont
    *           Flag identifying that the referenced font must default to the default fixed font, 
    *           if is not specified.
    */
   public GuiFontResolver(Window<?> window, GuiDriver gd, BaseConfig config, boolean fixedFont)
   {
      this.window = window;
      this.gd = gd;
      this.config = config;
      this.fixedFont = fixedFont;
   }
   
   /**
    * Resolve font.
    */
   public FontCache font()
   {
      int font = resolveFontNum();

      FontCache cached = fontCache.get(font);
      
      if (cached == null)
      {
         fontChanged = true;
         FontDetails<?> guiFont = FontManager.resolveFont(window, font, gd);
         
         int legacyHeight = FontManager.getFontHeight(window, font, gd);
         int legacyWidth = FontManager.getFontWidth(window, font, gd);
         
         // font is already scaled when is created, no reason to scale it again
         cached = new FontCache(legacyHeight, guiFont);
         fontCache.put(font, cached);
      }
      else
      {
         fontChanged = false;
      }
      
      return cached;
   }
   
   /**
    * Resolve the font number, in case the configuration has an undefined font.
    * 
    * @return   If the {@link #config configured} font is undefined, then either the 
    *           {@link FontManager#DEFAULT_FONT} or {@link FontManager#DEFAULT_FIXED_FONT} is 
    *           returned, depending on the {@link #fixedFont} flag.
    */
   public int resolveFontNum()
   {
      int font = config.font;
      
      if (config.font == FontManager.UNDEFINED_FONT && config.frameId != -1)
      {
         // get it from the frame
         Frame<?> w = ThinClient.getInstance().getFrame(config.frameId);
         
         if (w != null && w.config() != null)
         {
            font = w.config().font;
         }
      }
      
      return FontManager.resolveFont(font, fixedFont);
   }

   /**
    * Font cache item.
    */
   public static class FontCache
   {
      /** Font height. */
      public final int fontHeight;
      
      /** Font. */
      public final FontDetails<?> font;
      
      /**
       * Constructor.
       * 
       * @param    fontSize
       *           Font size.
       * @param    font
       *           Font instance.
       */
      public FontCache(int fontSize, FontDetails<?> font)
      {
         this.fontHeight = fontSize;
         this.font = font;
      }
   }      
}
