/*
** Module   : MouseHandler.java
** Abstract : Provides mouse event support for the swing GUI driver.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.
** 002 CA  20141021 Improved support for mouse events: fixed cases when the widgets need to be 
**                  notified the mouse has exited/entered it and cases when a widget needs to be
**                  notified of any mouse event (i.e. to hide a popup menu).
** 003 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 004 EVL 20141222 Adding methods to extract tooltip handlers from action map.
** 005 EVL 20150106 Fixed javadoc error.
** 006 HC  20150323 Removed business logic from config classes.
** 007 GES 20150501 Moved to a more generic package.
*/

package com.goldencode.p2j.ui.client.gui.driver;

// this usage of AWT events/Component is intentional, they are being used separately from the
// AWT/Swing event model; no other AWT/Swing usage is occurring in this class
import java.awt.Component;
import java.awt.event.*;

import java.util.*;

import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Provide mouse event support for the swing GUI driver.
 */
class MouseHandler
extends MouseAdapter
{
   /** Context-local data for this mouse-handler. */
   private static final ContextLocal<WorkArea> local = new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      {
         return new WorkArea();
      }
   };

   /** The GUI window emulator with which this mouse handler is registered. */
   private final EmulatedWindowState ews;
   
   /** The state of mouse event capturing: when {@code false}, mouse events are ignored. */
   private boolean state = false;
   
   /** Access to {@link ThinClient}. */
   private ThinClient tc;
   
   /**
    * Create a new mouse handler, for the specified window emulator.
    * 
    * @param    ews
    *           The window emulator.
    */
   public MouseHandler(EmulatedWindowState ews)
   {
      this.ews = ews;
      this.tc = ThinClient.getInstance();
   }

   /**
    * Notify the widget of any mouse events, generated after the given timestamp (in millis).
    * 
    * @param    wid
    *           The widget to be notified.
    * @param    when
    *           Only events generated after this timestamp will be notified.
    */
   public static void listenAnyMouseEvent(int wid, long when)
   {
      WorkArea wa = local.get();
      
      wa.anyActions.put(wid, when);
   }
   
   /**
    * Register the specified mouse action with the specified widget.
    * 
    * @param    wid
    *           The widget ID.
    * @param    action
    *           The action to be triggered for this widget, via mouse.
    */
   public static void registerWidgetAction(int wid, MouseWidgetAction action)
   {
      WorkArea wa = local.get();
      
      List<MouseWidgetAction> actions = wa.actions.get(wid);
      if (actions == null)
      {
         wa.actions.put(wid, actions = new ArrayList<>());
      }
            
      actions.add(action);
   }
   
   /**
    * Deregister the specified widget for mouse events.
    * 
    * @param    wid
    *           The widget ID.
    */
   public static void deregisterWidget(int wid)
   {
      WorkArea wa = local.get();
      
      wa.actions.remove(wid);
      wa.anyActions.remove(wid);
   }
   
   /**
    * Getting the list of the widgets with registered tooltip workers.
    *           
    * @return   The widget list for ones that have the registered worker.
    */
   public static Integer[] getWidgetListForTooltipWorkers()
   {
      // the resuls will be here
      ArrayList<Integer> widLst = new ArrayList<Integer>();

      WorkArea wa = local.get();
      // enumerate actions map
      Iterator<Integer> iter = wa.actions.keySet().iterator();
      while (iter.hasNext())
      {
         // get the widget ID
         Integer wId = iter.next();
         // get the list of the actions for this widget
         List<MouseWidgetAction> handlers = wa.actions.get(wId);
         // look the handler array for tooltip worker
         if (handlers != null)
         {
            // walk through handlers
            Iterator<MouseWidgetAction> iterAction = handlers.iterator();
            while (iterAction.hasNext())
            {
               if (iterAction.next() instanceof MouseTooltipWorker)
               {
                  // add widget ID to result list 
                  widLst.add(wId);
                  // only one tooltip worker per widget, go to next widget ID
                  break;
               }
            }
         }
      }
      
      return widLst.toArray(new Integer[widLst.size()]);
   }
   
   /**
    * Deregisters the specified widget for tooltip worker events.  The returned worker object
    * should be stored somewhere in caller to be registered further.
    * 
    * @param    wId
    *           The widget ID.
    *           
    * @return   The previously registered tooltip worker for given widgwt or <code>null</code> if
    *           the tooltip worker was not registered for this widget.
    */
   public static MouseWidgetAction deregisterTooltipWorker(int wId)
   {
      WorkArea wa = local.get();
      
      // get the list of the actions for this widget
      List<MouseWidgetAction> handlers = wa.actions.get(wId);
      // look the handler for tooltip worker
      if (handlers != null)
      {
         // walk through handlers
         Iterator<MouseWidgetAction> iterAction = handlers.iterator();
         while (iterAction.hasNext())
         {
            MouseWidgetAction action = iterAction.next(); 
            // found the tooltip handler
            if (action instanceof MouseTooltipWorker)
            {
               // remove handler from the list, no more tooltip events will be worked for
               // this widget
               iterAction.remove();
               // the handler removed, return copy to the caller, only one tooltip handler
               // per widget possible
               return action;
            }
         }
      }
      
      // if we are here - tooltip worker not found for the given widget
      return null;
   }
   
   /**
    * Handle the specified event, in the context of the specified widget.
    * 
    * @param    sourceID
    *           The source widget ID.
    * @param    evt
    *           The mouse event.
    */
   public static void handleMouseEvent(int sourceID, final MouseEvent evt)
   {
      @SuppressWarnings("unchecked")
      Widget<GuiOutputManager> source = 
         (Widget<GuiOutputManager>) OutputManager.resolveWidget(sourceID);
      
      if (source == null)
      {
         System.err.println("Received mouse event for zombie source " + sourceID + ": " + 
                            evt.toString());
         return;
      }
      
      applyMouseEvent(source, evt, true);
      
      ThinClient.getInstance().eventDrawingBracket(source, false, true, new Runnable()
      {
         @Override
         public void run()
         {
            processAnywhereActions(evt);
         }
      });

      processWidgetActions(sourceID, evt);
   }
   
   /**
    * Notify all widgets listening for any mouse events.
    * 
    * @param   evt
    *          The mouse event.
    */
   private static void processAnywhereActions(MouseEvent evt)
   {
      WorkArea wa = local.get();

      for (int wid : wa.anyActions.keySet())
      {
         long regTime = wa.anyActions.get(wid);
         long evtTime = evt.getWhen();
         
         if (evtTime <= regTime)
         {
            continue;
         }

         @SuppressWarnings("unchecked")
         Widget<GuiOutputManager> source = 
            (Widget<GuiOutputManager>) OutputManager.resolveWidget(wid);
         
         applyMouseEvent(source, evt, false);
      }
   }
   
   /**
    * Process the mouse actions registered for the specified source.
    * 
    * @param    sourceID
    *           The source widget ID.
    * @param    evt
    *           The mouse event.
    */
   private static void processWidgetActions(int sourceID, MouseEvent evt)
   {
      WorkArea wa = local.get();

      List<MouseWidgetAction> actions = wa.actions.get(sourceID);
      
      if (actions != null)
      {
         for (MouseWidgetAction action : actions)
         {
            switch (evt.getID())
            {
               case MouseEvent.MOUSE_CLICKED:
                  action.mouseClicked(evt);
                  break;
               
               case MouseEvent.MOUSE_PRESSED:
                  action.mousePressed(evt);
                  break;
               
               case MouseEvent.MOUSE_RELEASED:
                  action.mouseReleased(evt);
                  break;
               
               case MouseEvent.MOUSE_ENTERED:
                  action.mouseEntered(evt);
                  break;
               
               case MouseEvent.MOUSE_EXITED:
                  action.mouseExited(evt);
                  break;
               
               case MouseEvent.MOUSE_DRAGGED:
                  action.mouseDragged(evt);
                  break;
               
               case MouseEvent.MOUSE_MOVED:
                  action.mouseMoved(evt);
                  break;
               
               case MouseEvent.MOUSE_WHEEL:
                  action.mouseWheelMoved((MouseWheelEvent) evt);
                  break;
            }
         }
      }
   }
   
   /**
    * Apply the specified mouse event to the specified source.
    * 
    * @param    mouseSource
    *           The mouse source.
    * @param    evt
    *           The mouse event.
    */
   private static void applyMouseEvent(Widget<GuiOutputManager> mouseSource,
                                       MouseEvent               evt,
                                       boolean                  notifyMove)
   {
      WorkArea wa = local.get();
      
      switch (evt.getID())
      {
         case MouseEvent.MOUSE_CLICKED:
            mouseSource.mouseClicked(evt);
            break;
         
         case MouseEvent.MOUSE_PRESSED:
            mouseSource.mousePressed(evt);
            break;
         
         case MouseEvent.MOUSE_RELEASED:
            mouseSource.mouseReleased(evt);
            break;
         
         case MouseEvent.MOUSE_ENTERED:
            wa.lastHoveredWidget = mouseSource;
            
            mouseSource.mouseEntered(evt);
            break;
         
         case MouseEvent.MOUSE_EXITED:
            wa.lastHoveredWidget = null;
            
            mouseSource.mouseExited(evt);
            break;
         
         case MouseEvent.MOUSE_DRAGGED:
            mouseSource.mouseDragged(evt);
            break;
         
         case MouseEvent.MOUSE_MOVED:
            // if the widget has changed, send EXITED/ENTERED events
            if (notifyMove && wa.lastHoveredWidget != null && wa.lastHoveredWidget != mouseSource)
            {
               MouseEvent exitEvt = new MouseEvent((Component) evt.getSource(), 
                                                   MouseEvent.MOUSE_EXITED, 
                                                   evt.getWhen(), 
                                                   evt.getModifiers(), 
                                                   evt.getX(), 
                                                   evt.getY(), 
                                                   evt.getClickCount(), 
                                                   false, 
                                                   evt.getButton());
               wa.lastHoveredWidget.mouseExited(exitEvt);
               processWidgetActions(wa.lastHoveredWidget.getId().asInt(), exitEvt);
               
               MouseEvent enterEvt = new MouseEvent((Component) evt.getSource(),
                                                    MouseEvent.MOUSE_ENTERED, 
                                                    evt.getWhen(),
                                                    evt.getModifiers(),
                                                    evt.getX(),
                                                    evt.getY(),
                                                    evt.getClickCount(),
                                                    false,
                                                    evt.getButton());
               mouseSource.mouseEntered(enterEvt);
               processWidgetActions(mouseSource.getId().asInt(), enterEvt);
               
               wa.lastHoveredWidget = mouseSource;
            }
            else
            {
               mouseSource.mouseMoved(evt);
            }
            break;
         
         case MouseEvent.MOUSE_WHEEL:
            mouseSource.mouseWheelMoved((MouseWheelEvent) evt);
            break;
      }
   }
   
   /**
    * Enable or disable capture of mouse events.
    * 
    * @param    state
    *           {@code true} to capture mouse events, {@code false} to ignore them.
    *           
    * @return   The previous state.
    */
   public boolean captureMouseEvents(boolean state)
   {
      boolean prev = this.state;
      
      this.state = state;
      
      return prev;
   }

   /**
    * Notification of a mouse clicked event.  Event will be posted via 
    * {@link ThinClient#postMouseEvent} only if the {@link #state} flag is set to {@code true}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      if (state)
      {
         tc.postMouseEvent(e, ews.getWindowId());
      }
   }

   /**
    * Notification of a mouse pressed event.  Event will be posted via 
    * {@link ThinClient#postMouseEvent} only if the {@link #state} flag is set to {@code true}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      if (state)
      {
         tc.postMouseEvent(e, ews.getWindowId());
      }
   }
   
   /**
    * Notification of a mouse dragged event.  Event will be posted via 
    * {@link ThinClient#postMouseEvent} only if the {@link #state} flag is set to {@code true}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      if (state)
      {
         tc.postMouseEvent(e, ews.getWindowId());
      }
   }
   
   /**
    * Notification of a mouse released event.  Event will be posted via 
    * {@link ThinClient#postMouseEvent} only if the {@link #state} flag is set to {@code true}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      if (state)
      {
         tc.postMouseEvent(e, ews.getWindowId());
      }
   }
   
   /**
    * Notification of a mouse exited event.  Event will be posted via 
    * {@link ThinClient#postMouseEvent} only if the {@link #state} flag is set to {@code true}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseExited(MouseEvent e)
   {
      if (state)
      {
         tc.postMouseEvent(e, ews.getWindowId());
      }
   }
   
   /**
    * Notification of a mouse entered event.  Event will be posted via 
    * {@link ThinClient#postMouseEvent} only if the {@link #state} flag is set to {@code true}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseEntered(MouseEvent e)
   {
      if (state)
      {
         tc.postMouseEvent(e, ews.getWindowId());
      }
   }
   
   /**
    * Notification of a mouse moved event.  Event will be posted via 
    * {@link ThinClient#postMouseEvent} only if the {@link #state} flag is set to {@code true}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseMoved(MouseEvent e)
   {
      if (state)
      {
         tc.postMouseEvent(e, ews.getWindowId());
      }
   }
   
   /**
    * Notification of a mouse wheel moved event.  Event will be posted via 
    * {@link ThinClient#postMouseEvent} only if the {@link #state} flag is set to {@code true}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseWheelMoved(MouseWheelEvent e)
   {
      if (state)
      {
         tc.postMouseEvent(e, ews.getWindowId());
      }
   }
   
   /**
    * Container for mouse-related data.
    */
   private static class WorkArea
   {
      /**
       * Map of widgets listening for any mouse action.  Only events after the specified 
       * timestamp will be sent to the widget. 
       */
      private Map<Integer, Long> anyActions = new HashMap<>();

      /** Registered widgets which react on mouse events. */
      private Map<Integer, List<MouseWidgetAction>> actions = new HashMap<>();

      /** The last hovered widget. */
      private Widget<GuiOutputManager> lastHoveredWidget = null;
   }
}
