/*
** Module   : MousePopupable.java
** Abstract : Implementation of a mouse action which shows a popup.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.
** 002 CA  20141021 Better implementation of the mouse popupable action.
** 003 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 004 HC  20150323 WidgetId.getId() renamed to asInt().
** 005 GES 20150501 Moved to a more generic package.
*/

package com.goldencode.p2j.ui.client.gui.driver;

// this usage of AWT events is intentional, they are being used separately from the AWT/Swing
// event model; no other AWT/Swing usage is occurring in this class
import java.awt.event.*;

import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;

/**
 * Listener for mouse actions which can trigger a popup to be shown for a specified widget.
 */
public class MousePopupable
extends MouseWidgetAction
{
   /** The button on which the popup is associated. */
   private final int btn;

   /** The popup associated with this widget. */
   private PopupMenu popup;

   /**
    * Create a new mouse popupable action, for the specified widget.
    * 
    * @param    wid
    *           The widget id.
    * @param    btn
    *           The button to be pressed, when showing the popup.
    * @param    pid
    *           The popup widget ID.
    */
   public MousePopupable(int wid, int btn, int pid)
   {
      super(wid);
      
      this.popup = (PopupMenu) OutputManager.resolveWidget(pid);

      this.btn = btn;
   }

   /**
    * Notification of a mouse clicked event for this {@link MouseWidgetAction#widget}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      drawPopup(e);
   }

   /**
    * Notification of a mouse pressed event for this {@link MouseWidgetAction#widget}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      drawPopup(e);
   }
   
   /**
    * Draw the {@link #popup}, based on current popup state and mouse event.
    * 
    * @param    evt
    *           The current mouse event.
    */
   private void drawPopup(MouseEvent evt)
   {
      WindowGuiImpl window = (WindowGuiImpl) widget.ancestor();
      gd.selectWindow(window.config().id.asInt());

      if (popup.isVisible())
      {
         ThinClient.getInstance().eventDrawingBracket(window, false, true, new Runnable()
         {
            @Override
            public void run()
            {
               popup.hide();
            }
         });
         
         window.getContentPane().moveToBottom(popup);
      }

      int evtID = evt.getID();
      if (evt.getButton() == btn && evtID == MouseEvent.MOUSE_CLICKED)
      {
         window.getContentPane().moveToTop(popup);
         
         int nx = Math.min(evt.getX(), 
                          window.physicalDimension().width - popup.physicalDimension().width);
         int ny = evt.getY();
         
         double x = window.screen().coordinates().columnFromNative(nx);
         double y = window.screen().coordinates().rowFromNative(ny);
         
         popup.setLocation(x, y);
         popup.setPhysicalLocation(nx, ny);
         popup.doLayout();
         popup.setVisible(true);

         gd.listenAnyMouseEvent(popup.config().id.asInt(), evt.getWhen() + 1);
      }
   }
}
